import ode
from math import *
from gamelib.util import make_id
from gamelib.event import Event
x,y,z,w = (1,2,3,0)

CATAGORY_GENERIC = int('00000000000000000000000000000001', 2)
CATAGORY_GEOM = int('00000000000000000000000000000010', 2)

def quat_get_euler(quat):
    if type(quat) != type([]):
        quat = list(quat)

    t = quat[x] * quat[y] + quat[z] * quat[w]

    if t > 0.4999:
        heading = 2 * atan2(quat[x], quat[w])
        attitude = pi / 2
        bank = 0
    elif t < -0.4999:
        heading = -2 * atan2(quat[x], quat[w])
        attitude = -pi / 2
        bank = 0
    else:
        sqx = quat[x] ** 2
        sqy = quat[y] ** 2
        sqz = quat[z] ** 2
        heading = atan2(2 * quat[y] * quat[w] - 2 * quat[x] * quat[z],
                             1 - 2 * sqy - 2 * sqz)
        attitude = asin(2 * t)
        bank = atan2(2 * quat[x] * quat[w] - 2 * quat[y] * quat[z],
                          1 - 2 * sqx - 2 * sqz)

    return heading, attitude, bank


def calcRotMatrix(axis, angle):
    """
    Returns the row-major 3x3 rotation matrix defining a rotation around axis by
    angle.
    """
    cosTheta = cos(angle)
    sinTheta = sin(angle)
    t = 1.0 - cosTheta
    return (
        t * axis[0]**2 + cosTheta,
        t * axis[0] * axis[1] - sinTheta * axis[2],
        t * axis[0] * axis[2] + sinTheta * axis[1],
        t * axis[0] * axis[1] + sinTheta * axis[2],
        t * axis[1]**2 + cosTheta,
        t * axis[1] * axis[2] - sinTheta * axis[0],
        t * axis[0] * axis[2] - sinTheta * axis[1],
        t * axis[1] * axis[2] + sinTheta * axis[0],
        t * axis[2]**2 + cosTheta)


def quat_set(angle, x, y, z):

    angle = radians(angle + 90.0)
    # normalize
    temp = x*x + y*y + z*z
    dist = 1.0 / sqrt(temp)
    x *= dist
    y *= dist
    z *= dist

    w = cos(angle / 2.0)

    return (w, x, y, z)


class Simulation:

    objects = {}
    stepped_objects = {}
    post_step_objects = {}
    joints = []

    def __init__(self):
        self.world = ode.World()
        self.world.setGravity((0.0, -9.81, 0.0))
        self.world.setERP(0.6)
        self.world.setCFM(1E-5)

        self.world.setAutoDisableFlag(True)
        self.world.setAutoDisableLinearThreshold(0.2)
        self.world.setAutoDisableAngularThreshold(0.2)
        self.world.setAutoDisableSteps(5)
        self.world.setAutoDisableTime(0.25)

        self.contact_group = ode.JointGroup()

        self.space = ode.QuadTreeSpace((0.0, 0.0, 0.0), (10.0, 10.0, 10.0), 4)

        self.objects = {}
        self.stepped_objects = {}
        self.post_step_objects = {}
        self.joints = []
        
        self.cleanup_list = []

    def get_world_space(self):
        return (self.world, self.space)
    
    def remove_object(self, dynamics_id):
        self.cleanup_list.append(dynamics_id)

    def step(self):

        for id, object in self.stepped_objects.iteritems():
            object.step()

        self.space.collide((), self.collision_callback)
        self.world.quickStep(1.0/30.0)
        self.contact_group.empty()

        for id, object in self.post_step_objects.iteritems():
            object.post_step()
        
        for dyn_id in self.cleanup_list:
            obj=None
            if dyn_id in self.objects.keys():
                obj = self.objects[dyn_id]
                del self.objects[dyn_id]
            if dyn_id in self.stepped_objects.keys():
                obj = self.stepped_objects[dyn_id]
                del self.stepped_objects[dyn_id]
            if dyn_id in self.post_step_objects.keys():
                obj = self.post_step_objects[dyn_id]
                del self.post_step_objects[dyn_id]
            self.space.remove(obj.dynamics_geom)
            del obj
        self.cleanup_list = []


    def add_object(self, object):

        store = self.objects if not object.step_me else self.stepped_objects

        id = make_id(store=store)
        object.create_self(self, id)
        store[id] = object

        if callable(getattr(object, "post_step", False)):
            self.post_step_objects[id] = object

        return id


    def collision_callback(self, args, geom1, geom2):

        contacts = ode.collide(geom1, geom2)
        min_bounce = 0.0
        bounce = 0.0
        friction = 0.0

        if not len(contacts):
            return

        # the contact information generated by ray collisions
        # is not for contact joints
        make_contacts = True
        if type(geom1) is ode.GeomRay:
            make_contacts = False

        if type(geom2) is ode.GeomRay:
            make_contacts = False

        if not geom1.super_object.collide or not geom2.super_object.collide:
            make_contacts = False

        # Ignore player and playercopy collisions
        if geom1.super_object.isplayer and geom2.super_object.isplayer:
            make_contacts = False
        
        #print geom1, geom2
        touching_platform = False
        if geom1.super_object.isplatform:
            body = geom2.getBody()
            if body:
                # Mitigates the rope thing somewhat, not completely but it works... kinda ;)
                if getattr(geom2.super_object, "touching_rope", False) and\
                   geom2.super_object.dyn_get_position()[1] - geom2.super_object.dynamics_size[1]/2.0 <= geom1.getPosition()[1]:
                    make_contacts = False
                    touching_platform = True


                if body.getLinearVel()[1] >= 0.8:
                    make_contacts = False
                    touching_platform = True


        if geom2.super_object.isplatform:
            body = geom1.getBody()
            if body:
                if getattr(geom1.super_object, "touching_rope", False) and\
                   geom1.super_object.dyn_get_position()[1] - geom1.super_object.dynamics_size[1]/2.0 <= geom2.getPosition()[1]:
                    make_contacts = False
                    touching_platform = True

                if body.getLinearVel()[1] >= 0.8:
                    make_contacts = False
                    touching_platform = True
        
        # Let copies pass through harmful objects
        if getattr(geom1.super_object, "isghost", False) and geom2.super_object.isharmful:
            make_contacts = False
        if getattr(geom2.super_object, "isghost", False) and geom1.super_object.isharmful:
            make_contacts = False

        #print geom1.super_object.collide
        #print geom2.super_object.collide

        for (index, contact) in enumerate(contacts):

            if index == 0:
                if make_contacts:

                    body1 = geom1.getBody()
                    body2 = geom2.getBody()
                    contact_info = contact.getContactGeomParams()

                    if body1:
                        if contact_info[2] > 0.01:
                            geom1.super_object.dyn_collide(geom2.super_object, contact_info)

                    if body2:
                        if contact_info[2] > 0.01:
                            geom2.super_object.dyn_collide(geom1.super_object, contact_info)

                    # Find the bounce for this collision. Take the higher bounce value
                    mat1 = geom1.super_object.dyn_material
                    mat2 = geom1.super_object.dyn_material

                    # Average them?
                    bounce = (mat1['bounce'] + mat2['bounce']) / 2.0
                    friction = (mat1['friction'] + mat2['friction']) / 2.0
                    min_bounce = (mat1['min_bounce'] + mat2['min_bounce']) / 2.0

                elif (geom1.super_object is not geom2.super_object):

                    contact_info = contact.getContactGeomParams()

                    if type(geom1) is ode.GeomRay:
                        geom1.super_object.is_colliding = True
                    else:
                        if not touching_platform:
                            geom1.super_object.dyn_collide(geom2.super_object, contact_info)

                    if type(geom2) is ode.GeomRay:
                        geom2.super_object.is_colliding = True
                    else:
                        if not touching_platform:
                            geom2.super_object.dyn_collide(geom1.super_object, contact_info)

            if make_contacts:
                contact.setBounce(bounce)
                contact.setBounceVel(min_bounce)
                contact.setMu(friction)
                j = ode.ContactJoint(self.world, self.contact_group, contact)
                j.attach(body1, body2)


class DynamicObject(Event):

    dynamics_size = (1.0, 1.0)
    dynamics_pos = (0.0, 0.0)
    dynamics_bound = 'box'

    dynamics_autodisable = True
    is_colliding = False

    catagory = CATAGORY_GEOM
    collides = CATAGORY_GENERIC

    isplatform = False

    isplayer = False


    id = 0
    collide = True

    dynamics_material = {'density': 255, 'bounce':0.0, 'bounce_min':0.0, 'friction':0.0 }
    is_colliding = False

    islift = False

    isharmful = False

    def __init__(self, pos=(0.0, 0.0)):
        Event.__init__(self)
        self.dynamics_pos = pos
        self.is_colliding = False


    def create_self(self, world, myid):
        self.dynamics_id = myid
        self.dynamics_world, self.dynamics_space = world.get_world_space()

        if self.dynamics_bound == 'sphere':
            radius = self.dynamics_size[0] * 0.5
            self.dynamics_geom = ode.GeomSphere(self.dynamics_space, radius)

        elif self.dynamics_bound == 'box':
            width, height = self.dynamics_size
            depth = 5.0

            self.dynamics_geom = ode.GeomBox(self.dynamics_space, lengths=(width, height, depth))

        elif self.dynamics_bound == 'capsule':
            radius, length = self.dynamics_size
            radius *= 0.5
            length *= 0.5

            self.dynamics_geom = ode.GeomCapsule(self.dynamics_space, radius=radius, length=length)

        # finally we want to attach this object to our body so that we can
        # reference it back
        self.dynamics_geom.super_object = self

        # set our bits
        self.dynamics_geom.setCategoryBits(self.catagory)
        self.dynamics_geom.setCollideBits(self.collides)


    def dyn_set_position(self, x, y):
        self.dynamics_geom.setPosition((x, y, 0.0))

    def dyn_get_position(self):
        x, y, z = self.dynamics_geom.getPosition()
        return x, y

    def dyn_get_rotation(self):
        return degrees(quat_get_euler(self.dynamics_geom.getQuaternion())[2])

    def dyn_get_rotation_raw(self):
        return self.dynamics_geom.getRotation()

    def dyn_set_rotation(self, angle):
        """ sets the rotation of the dynamic body """
        raise Exception, 'not implimentated dyn_set_rotation'


    # we do this so that the duck typing works nicely
    get_position = dyn_get_position
    get_rotation = dyn_get_rotation
    set_position = dyn_set_position
    set_rotation = dyn_set_rotation
    get_rotation_raw = dyn_get_rotation_raw

    def dyn_collide(self, object, info):
        self.emit('collide', object, info)




class DynamicBody(DynamicObject):

    dynamics_joints = []

    catagory = CATAGORY_GENERIC
    collides = CATAGORY_GENERIC | CATAGORY_GEOM

    def __init__(self, pos=(0.0, 0.0)):
        DynamicObject.__init__(self, pos)

    def create_self(self, world, myid):
        world.get_world_space()[0].setAutoDisableFlag(self.dynamics_autodisable)

        DynamicObject.create_self(self, world, myid)

        # we have our geometry, hopefully


        density = self.dynamics_material['density']
        self.dynamics_body = ode.Body(self.dynamics_world)
        self.dynamics_mass = ode.Mass()


        if self.dynamics_bound == 'sphere':
            self.dynamics_mass.setSphere(density, self.dynamics_size[0] * 0.5)

        elif self.dynamics_bound == 'box':
            width, height = self.dynamics_size
            depth = 5.0
            self.dynamics_mass.setBox(density, width, height, depth)

        elif self.dynamics_bound == 'capsule':
            radius, length = self.dynamics_size
            radius *= 0.5

            self.dynamics_mass.setCappedCylinder(density, 2, radius, length)



        self.dynamics_body.setMass(self.dynamics_mass)

        # create a 2d joint to keep things 2d
        joint = ode.Plane2DJoint(self.dynamics_world)
        joint.attach(self.dynamics_body, None)

        #self.dyn_world.setAutoDisableFlag(True)

        self.dynamics_joints.append(joint)

        self.dynamics_geom.setBody(self.dynamics_body)

        # finally we want to attach this object to our body so that we can
        # reference it back
        self.dynamics_body.super_object = self
        self.dynamics_geom.super_object = self

        world.get_world_space()[0].setAutoDisableFlag(True)


    def dyn_set_position(self, x, y):
        self.dynamics_body.setPosition((x, y, 0.0))

    def dyn_get_position(self):
        x, y, z = self.dynamics_body.getPosition()
        return x, y

    def dyn_get_rotation(self):
        return degrees(quat_get_euler(self.dynamics_body.getQuaternion())[2])

    def dyn_get_rotation_raw(self):
        return self.dynamics_body.getRotation()

    def dyn_set_rotation(self, angle):
        """ sets the rotation of the dynamic body """
        raise Exception, 'not implimentated dyn_set_rotation'


    # we do this so that the duck typing works nicely
    get_position = dyn_get_position
    get_rotation = dyn_get_rotation
    set_position = dyn_set_position
    set_rotation = dyn_set_rotation
    get_rotation_raw = dyn_get_rotation_raw

    def step(self):
        pass


class DynamicBodyUpright(DynamicBody):

    dyn_bound_type = 'capsule'
    step_me = True
    dynamics_autodisable = False

    rot_mat = calcRotMatrix((1.0, 0.0, 0.0), 90)

    def __init__(self, pos=(0.0, 0.0)):
        DynamicBody.__init__(self)

    def step(self):
        # get current rotation

        self.dynamics_body.setAngularVel((0.0, 0.0, 0.0))
        self.dynamics_body.setRotation(self.rot_mat)
        #self.dynamics_body.setRotation((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))


    def dyn_get_rotation(self):
        return 0

    def dyn_get_rotation_raw(self):
        return (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)

    get_rotation = dyn_get_rotation
    get_rotation_raw = dyn_get_rotation_raw
